/*
 * This file is part of the TASCAR software, see <http://tascar.org/>
 *
 * Copyright (c) 2018 Giso Grimm
 * Copyright (c) 2020 Giso Grimm
 * Copyright (c) 2021 Giso Grimm
 */
/*
 * TASCAR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, version 3 of the License.
 *
 * TASCAR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHATABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License, version 3 for more details.
 *
 * You should have received a copy of the GNU General Public License,
 * Version 3 along with TASCAR. If not, see <http://www.gnu.org/licenses/>.
 */

#include "session.h"
#include <complex>

const std::complex<double> i(0.0, 1.0);

class mpu6050track_t : public TASCAR::actor_module_t {
public:
  mpu6050track_t(const TASCAR::module_cfg_t& cfg);
  ~mpu6050track_t();
  void configure();
  void update(uint32_t frame, bool running);
  static int osc_setrot(const char* path, const char* types, lo_arg** argv,
                        int argc, lo_message msg, void* user_data);
  static int osc_setrotypr(const char* path, const char* types, lo_arg** argv,
                           int argc, lo_message msg, void* user_data);
  static int osc_setrotgyr(const char* path, const char* types, lo_arg** argv,
                           int argc, lo_message msg, void* user_data);
  static int osc_calib(const char* path, const char* types, lo_arg** argv,
                       int argc, lo_message msg, void* user_data);
  static int osc_calib1(const char* path, const char* types, lo_arg** argv,
                        int argc, lo_message msg, void* user_data);
  void setrot(float* e);
  void setrotypr(float* e);
  void setrotgyr(double t, double* gyr, double* acc);
  void calib();
  void calib1(bool b_cal);

private:
  std::string id;
  std::string mode;
  std::string calib_start;
  std::string calib_end;
  uint32_t ypraxis;
  uint32_t gyraxis;
  uint32_t zaxis;
  uint32_t rotaxis;
  double rotscale;
  double zscale;
  double gyrscale;
  double scale;
  double tau;
  double tauz;
  std::complex<double> z_mean;
  double c;
  double cz;
  double rot;
  double zshift;
  double zshiftlp;
  bool b_calib;
  double tprev;
  lo_message msg_calib;
  bool use_translation;
};

void mpu6050track_t::configure()
{
  actor_module_t::configure();
  c = exp(-1.0 / (tau * f_fragment));
  cz = exp(-1.0 / (tauz * f_fragment));
}

int mpu6050track_t::osc_calib(const char*, const char*, lo_arg**, int,
                              lo_message, void* user_data)
{
  if(user_data)
    ((mpu6050track_t*)user_data)->calib();
  return 0;
}

int mpu6050track_t::osc_calib1(const char*, const char*, lo_arg** argv, int,
                               lo_message, void* user_data)
{
  if(user_data)
    ((mpu6050track_t*)user_data)->calib1(argv[0]->i);
  return 0;
}

int mpu6050track_t::osc_setrot(const char*, const char* types, lo_arg** argv,
                               int argc, lo_message, void* user_data)
{
  // 0 = time
  // 1 = rx
  // 2 = ry <- here up-axis
  // 3 = rz
  if(user_data && (argc == 4) && (types[2] == 'f')) {
    float data[3];
    data[0] = argv[3]->f;
    data[1] = argv[2]->f;
    data[2] = argv[1]->f;
    ((mpu6050track_t*)user_data)->setrot(data);
  }
  return 1;
}

int mpu6050track_t::osc_setrotypr(const char*, const char* types, lo_arg** argv,
                                  int argc, lo_message, void* user_data)
{
  // 0 = time
  // 1 = yaw
  // 2 = pitch
  // 3 = roll
  if(user_data && (argc == 4) && (types[2] == 'f')) {
    float data[3];
    data[0] = DEG2RAD * (argv[1]->f);
    data[1] = DEG2RAD * (argv[2]->f);
    data[2] = DEG2RAD * (argv[3]->f);
    ((mpu6050track_t*)user_data)->setrotypr(data);
  }
  return 0;
}

int mpu6050track_t::osc_setrotgyr(const char*, const char* types, lo_arg** argv,
                                  int argc, lo_message, void* user_data)
{
  // 0 = time
  // 1 = yaw
  // 2 = pitch
  // 3 = roll
  if(user_data && (argc == 7) && (types[0] == 'd')) {
    double data[3];
    double acc[3];
    data[0] = DEG2RAD * (argv[4]->d);
    data[1] = DEG2RAD * (argv[5]->d);
    data[2] = DEG2RAD * (argv[6]->d);
    acc[0] = argv[1]->i;
    acc[1] = argv[2]->i;
    acc[2] = argv[3]->i;
    ((mpu6050track_t*)user_data)->setrotgyr(argv[0]->d, data, acc);
  }
  return 0;
}

void mpu6050track_t::setrot(float* data)
{
  TASCAR::zyx_euler_t e(rotscale * data[rotaxis], 0, 0);
  TASCAR::pos_t x(0, 1, 0);
  x *= e;
  double az(x.azim());
  if((-4 < az) && (az < 4))
    rot = az;
}

void mpu6050track_t::setrotypr(float* e)
{
  rot = e[std::min(2u, ypraxis)];
}

void mpu6050track_t::setrotgyr(double t, double* gyr, double* acc)
{
  double g(gyr[std::min(2u, gyraxis)]);
  double dt(t - tprev);
  tprev = t;
  if((0.0 < dt) && (dt < 0.1)) {
    rot += gyrscale * dt * g;
  }
  if(zaxis < 3) {
    zshift = zscale * acc[zaxis] * 0.00006103515625;
  }
}

void mpu6050track_t::calib()
{
  b_calib = true;
}

void mpu6050track_t::calib1(bool bcal)
{
  b_calib = true;
  if(bcal) {
    if(!calib_start.empty())
      session->dispatch_data_message(calib_start.c_str(), msg_calib);
  } else {
    if(!calib_end.empty())
      session->dispatch_data_message(calib_end.c_str(), msg_calib);
  }
}

mpu6050track_t::mpu6050track_t(const TASCAR::module_cfg_t& cfg)
    : actor_module_t(cfg), ypraxis(0), gyraxis(1), zaxis(-1), rotaxis(1),
      rotscale(DEG2RAD), zscale(1), gyrscale(0.06), scale(1.0), tau(2.0),
      tauz(0.5), z_mean(1.0), c(0.999), rot(0), zshift(0), zshiftlp(0),
      b_calib(false), tprev(0), msg_calib(lo_message_new()),
      use_translation(true)
{
  lo_message_add_int32(msg_calib, 1);
  GET_ATTRIBUTE_(id);
  GET_ATTRIBUTE_(tau);
  GET_ATTRIBUTE_(tauz);
  GET_ATTRIBUTE_(scale);
  GET_ATTRIBUTE_(gyrscale);
  GET_ATTRIBUTE_(rotaxis);
  GET_ATTRIBUTE_BOOL_(use_translation);
  if(2 < rotaxis)
    throw TASCAR::ErrMsg("Invalid rotation axis (must be 0, 1 or 2).");
  GET_ATTRIBUTE_(rotscale);
  GET_ATTRIBUTE_(ypraxis);
  GET_ATTRIBUTE_(gyraxis);
  GET_ATTRIBUTE_(zaxis);
  GET_ATTRIBUTE_(zscale);
  GET_ATTRIBUTE_(calib_start);
  GET_ATTRIBUTE_(calib_end);
  GET_ATTRIBUTE_(mode);
  if(mode.empty())
    mode = "euler";
  if(mode == "euler")
    session->add_method(id + "/euler", "ffff", &mpu6050track_t::osc_setrot,
                        this);
  else if(mode == "ypr")
    session->add_method(id + "/ypr", "ffff", &mpu6050track_t::osc_setrotypr,
                        this);
  else if(mode == "gyr")
    session->add_method(id + "/raw", "diiiddd", &mpu6050track_t::osc_setrotgyr,
                        this);
  else if(mode == "rotnew")
    session->add_method("/rot", "ffff", &mpu6050track_t::osc_setrot, this);
  else
    throw TASCAR::ErrMsg("Unknown mode \"" + mode + "\".");
  session->add_method(id + "/calib", "fff", &mpu6050track_t::osc_calib, this);
  session->add_method(id + "/calibration", "i", &mpu6050track_t::osc_calib1,
                      this);
}

mpu6050track_t::~mpu6050track_t() {}

void mpu6050track_t::update(uint32_t, bool)
{
  std::complex<double> z(std::exp(i * rot));
  if(b_calib) {
    b_calib = false;
    z_mean = z;
  }
  z_mean = c * z_mean + (1.0 - c) * z;
  double az(std::arg(z * conj(z_mean)));
  // check for normal values by using a comparison (result is false for nan and
  // inf):
  if((-4 < az) && (az < 4))
    set_orientation(TASCAR::zyx_euler_t(scale * az, 0, 0));
  else
    set_orientation(TASCAR::zyx_euler_t(0, 0, 0));
  // z-shifting:
  zshiftlp = cz * zshiftlp + (1.0 - cz) * zshift;
  if(use_translation)
    set_location(TASCAR::pos_t(0, 0, zshiftlp));
}

REGISTER_MODULE(mpu6050track_t);

/*
 * Local Variables:
 * mode: c++
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * compile-command: "make -C .."
 * End:
 */
